// turnTableNode - This is a DIVERSE DSO which loads files specified 
// by an environment variable, DPF_TURNTABLE_FILES, and spins the
// corresponding objects around the Z axis going through their centers.
//
// Authors: 
//     John Kelso
//       Virginia Tech, http://www.diverse.vt.edu/, http://www.cave.vt.edu/
//
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//      
//    
//    Steve Satterfield (SAVG) and John Kelso provided 
//    the idea for the program
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

// example turntableNode dtkAugment DSO

class TurntableNode : public dtkAugment {
public:
  
  TurntableNode(dpf *app) ;
  ~TurntableNode(void) ;
  int preConfig(void);
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  pfGroup* mygroup;
  pfSphere* allSphere;
  pfMatrix* allMove;
  pfMatrix* allMoveBack;
  // dtkTime is a gettimeofday() wrapper
  dtkTime t;
  float lastRot;
  //  Shared memory segment for the spin rate
  dtkSharedMem* shm;
  pfMatrix dpfToWorldMat ;
  pfMatrix worldToDpfMat ;
  pfVec3 _axis;
} ;

// create object and set description.
// be sure to validate it!
TurntableNode::TurntableNode(dpf *app_in) :
  dtkAugment("turntableNode")
{
  setDescription("loads files specified by DPF_TURNTABLE_FILES"
		 " environment variable and spins them around"
		 " the Z axis going through their centers") ;
  app = app_in;
  lastRot = 0.f;
  validate();
  //  Check if there is a segment "spin" to control the spin rate
  shm = new dtkSharedMem(sizeof(float), "spin");
  if(!shm || shm->isInvalid())
    shm = NULL ;

} ;

// destructor
TurntableNode::~TurntableNode(void) {
  delete allSphere;
  delete allMove ;
  delete allMoveBack ;
}


int TurntableNode::preConfig(void) {
  
  // init file loaders
  int sufc = 0 ;
  char *suf ;
  char **sufv = NULL ;
  char *env = getenv("DPF_TURNTABLE_FILES") ;
  char *file ;
  
  if (env) {
    file = strdup(env) ;
    // strtok() is broken in GNU libc, so we roll our own here.
    char *end = file ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // load file, or NULL, here
      if (strlen(file)>0) {
        suf = strrchr(file, '.') ;
        if (suf) {
	  //printf("foo- suffix = \"%s\"\n",suf) ;
	  int suf_found = 0 ;
	  for (int c = 0; c<sufc; c++) {
	    //printf("foo- comparing %s with %s\n",suf,sufv[c]) ;
	    if (!strcmp(suf,sufv[c])) {
	      suf_found = 1 ;
	      break ;
	      
	    }
	  }
	  if (!suf_found) {
	    //printf("loading converter for \"%s\"\n",suf) ;
	    sufv = (char **) realloc(sufv, sizeof(char *)*(sufc+1));
	    sufv[sufc] = strdup(suf) ;
	    sufc++ ;
	    if(!pfdInitConverter(suf)) {
	      dtkMsg.add(DTKMSG_ERROR,
			 "pfdInitConverter(%s) failed.\n",
			 suf) ;
	      return DTK_CALLBACK_ERROR;
	    }
          }
        } else {
	  dtkMsg.add(DTKMSG_ERROR,
		     "No file suffix found for file \"%s\"\n",file) ;
	  return DTK_CALLBACK_ERROR;
        }
	
      }      
      file = end;
    }
  } 
  return DTK_REMOVE_CALLBACK;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int TurntableNode::postConfig(void) {
  t.reset();
  display = app->display();
  mygroup = new pfGroup();
  display->world()->addChild(mygroup);
  pfDCS* dcs ;
  
  char *old_path = (char *) pfGetFilePath();
  char *path = NULL;
  if(old_path)
    {
      // copy old_path from static memory to the heap.
      old_path = dtk_strdup(old_path);
      path = (char *) dtk_malloc(strlen(old_path) +
				 strlen("%s:.:"DPF_DATA_DIR));
      sprintf(path,"%s:.:"DPF_DATA_DIR, old_path);
      pfFilePath(path);
    }
  else // old_path == NULL
    pfFilePath(".:"DPF_DATA_DIR);
  
  // parse through the environment variable
  char *file = getenv("DPF_TURNTABLE_FILES") ;
  if (file) {
    // strtok() is broken in GNU libc, so we roll our own here.
    char *end = file ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // load file, or NULL, here
      if (strlen(file)>0) {
	dcs = new pfDCS();
	//printf("dcs=%p\n",dcs) ;
	mygroup->addChild(dcs);
	dcs->addChild(pfdLoadFile(file)) ;
      }      
      file = end;
    }
  } 
  // assign an array of bounding spheres for the objects
  allSphere = new pfSphere[mygroup->getNumChildren()];
  allMove = new pfMatrix[mygroup->getNumChildren()];
  allMoveBack = new pfMatrix[mygroup->getNumChildren()];

  for(int i = 0; i<mygroup->getNumChildren(); i++) {
    allSphere[i] = dpf_getBSphere(mygroup->getChild(i));
#if 0
    printf("center = (%f,%f,%f), radius = %f\n",
	   allSphere[i].center[PF_X], 
	   allSphere[i].center[PF_Y],       
	   allSphere[i].center[PF_Z],
	   allSphere[i].radius) ;
#endif
    allMove[i].makeTrans(allSphere[i].center[PF_X], 
			 allSphere[i].center[PF_Y],       
			 allSphere[i].center[PF_Z]) ;

    allMoveBack[i].makeTrans(-allSphere[i].center[PF_X], 
			     -allSphere[i].center[PF_Y],       
			     -allSphere[i].center[PF_Z]) ;

  }
  return DTK_REMOVE_CALLBACK ; 
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int TurntableNode::postFrame(void) {

  pfCoord c1;
  c1.xyz.set(0.0f, 0.0f, 1.0f);
  c1.hpr.set(0.0f, 0.0f, 0.0f);
  pfCoord c0;
  c0.xyz.set(0.0f, 0.0f, 0.0f);
  c0.hpr.set(0.0f, 0.0f, 0.0f);

  display->dpfToWorld(&c1);
  display->dpfToWorld(&c0);

  // hpr mode sets _axis based on wand hpr, in postFrame
  _axis = c1.xyz - c0.xyz;                                                  
  _axis.normalize();

  float delta_t = (float) t.delta();

  pfDCS *dcs ;
  float step ;
  if (shm)
    shm->read(&step);
  else
    step = 10.f ;
  //printf("step = %f               \r",step) ;
  //fflush(stdout) ;
  lastRot += delta_t * step;
  for(int i = 0; i<mygroup->getNumChildren(); i++) {
    dcs = (pfDCS *) mygroup->getChild(i) ;
    pfMatrix spin ;
    spin = allMove[i] ;
    spin.preRot(lastRot, _axis[0], _axis[1],
		  _axis[2], spin);
    spin = allMoveBack[i] * spin ;
    dcs->setMat(spin) ;
  }
  return DTK_CONTINUE ; 
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new TurntableNode(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
